/* VzImagination (C) 2018-2019 Stephane Charette <stephanecharette@gmail.com>
 * $Id: TestKMeans.cpp 2741 2019-04-09 17:04:20Z stephane $
 */

#include "KMeans.hpp"
#include "Intervals.hpp"
#include <gtest/gtest.h>


TEST(KMeans, KMeans)
{
	vz::VDouble data =
	{
		1.0, 1.1, 1.2, 1.3, 1.22,
		1.01, 1.02, 1.03, 1.04, 1.01,
		3.3, 3.4, 3.5, 3.6,
		3.31, 3.32,
		7.2, 7.3, 7.4, 7.5,
		7.21,
		8, 9, 10, 11,
	};
	vz::VSizet assignments;
	vz::VDouble centroids;

	size_t iterations = vz::k_means(data, 3, assignments, centroids);

	std::cout << "data points: " << data.size() << ", iterations: " << iterations << ", centroids: " << centroids.size() << std::endl;

	for (size_t idx = 0; idx < centroids.size(); idx++)
	{
		std::cout << "centroid #" << idx << " is located at: " << centroids[idx] << std::endl;
	}
	for (size_t idx = 0; idx < data.size(); idx++)
	{
		std::cout << "data point " << data[idx] << " was assigned to cluster #" << assignments[idx] << std::endl;
	}
}


TEST(KMeans, SimpleTest)
{
	// only 6 values with obvious clusters at 1.2 and 5.8

	vz::VDouble data = { 1.1, 1.2, 1.3, 5.7, 5.8, 5.9 };

	vz::VSizet assignments;
	vz::VDouble centroids;
	const size_t iterations = vz::k_means(data, 2, assignments, centroids);

	ASSERT_EQ(assignments.size(), 6);
	ASSERT_EQ(centroids.size(), 2);

	const vz::VSizet expected_assignments = { 0, 0, 0, 1, 1, 1 };
	const vz::VDouble expected_centroids = { 1.2, 5.8 };

	ASSERT_EQ(assignments, expected_assignments );
	ASSERT_EQ(centroids, expected_centroids);
	ASSERT_EQ(iterations, 2);
}


TEST(KMeans, VectorTest)
{
	vz::VDouble data =
	{
		1.1, 1.2, 1.3, 1.4, 1.5, 1.6,
		5.7, 5.8, 5.9,
		9.2, 9.4, 9.6, 9.8,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		30, 31, 32, 33, 34
	};

	vz::VVDouble results;
	vz::VDouble centroids;
	const size_t iterations = vz::k_means_vector(data, 5, results, centroids);

	ASSERT_EQ(results.size(), centroids.size());
	std::cout << "Iterations: " << iterations << ", number of vectors: " << results.size() << std::endl;

	for (size_t idx = 0; idx < results.size(); idx++)
	{
		std::cout << "#" << idx << ": centroid=" << centroids.at(idx) << ", data=";
		for (auto d : results.at(idx))
		{
			std::cout << " " << d;
		}
		std::cout << std::endl;
	}
}


TEST(Intervals, Empty)
{
	vz::VDouble data;
	auto results = vz::intervals(data);
	ASSERT_EQ(results.size(), 0);
}


TEST(Intervals, InvalidThreshold)
{
	vz::VDouble data = {5.0};
	ASSERT_THROW(vz::intervals(data, 0), std::exception);

	ASSERT_THROW(vz::intervals(data, -1.0), std::exception);
}


TEST(Intervals, Simple1)
{
	vz::VDouble data = {5.0};
	auto results = vz::intervals(data);
	ASSERT_EQ(results.size(), 1);
	ASSERT_EQ(results[0].size(), 1);
	ASSERT_EQ(results[0][0], 5.0);
}


TEST(Intervals, Simple2)
{
	vz::VDouble data = {101, 102, 103, 104, 105};
	auto results = vz::intervals(data);
	ASSERT_EQ(results.size(), 1);
	ASSERT_EQ(results[0].size(), 5);
	ASSERT_EQ(results[0], data);
}


TEST(Intervals, Complex1)
{
	vz::VDouble data =
	{
		101, 102, 103, 104, 105, 106, 107, 108, 109,
		1.01, 1.02, 1.03, 1.04, 1.05
	};

	ASSERT_EQ(data.at(0), 101.0); // prior to sorting
	auto results = vz::intervals(data);
	ASSERT_EQ(data.at(0), 1.01); // after sorting

	std::cout << vz::to_string(results) << std::endl;

	ASSERT_EQ(results.size(), 2);
	ASSERT_EQ(results.at(0).size(), 5);
	ASSERT_EQ(results.at(1).size(), 9);

	const vz::VDouble v1 = {1.01, 1.02, 1.03, 1.04, 1.05};
	const vz::VDouble v2 = {101, 102, 103, 104, 105, 106, 107, 108, 109};
	ASSERT_EQ(results.at(0), v1);
	ASSERT_EQ(results.at(1), v2);
}


TEST(Intervals, Complex2)
{
	vz::VDouble data =
	{
		30, 31, 32, 33, 34,
		5.7, 5.8, 5.9,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		9.2, 9.4, 9.6, 9.8,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		1.1, 1.2, 1.3, 1.4, 1.5, 1.6,
		15.01, 15.02, 15.03, 15.04, 15.05, 15.06, 15.07, 15.08, 15.09,
		99
	};

	const vz::VVDouble results = vz::intervals(data);

	std::cout << vz::to_string(results) << std::endl;
}
